/*
 * Copyright (c) 1998, Larry Lile
 * All rights reserved.
 *
 * For latest sources and information on this driver, please
 * go to http://anarchy.stdio.com.
 *
 * Questions, comments or suggestions should be directed to
 * Larry Lile <lile@stdio.com>.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: if_tok.c,v 1.19 1999/01/09 22:45:58 lile Exp $
 *
 * Much code stolen from many ethernet drivers too numerous to mention :)
 */

/*
 *
 * IBM Shared Ram and TROPIC Token-Ring
 *
 *   Token-Ring Network PC Adapter
 *   Token-Ring Network PC Adapter II
 *   Token-Ring Network Adapter/A
 *   Token-Ring Network 16/4 Adapter/A (long card)
 *   Token-Ring Network 16/4 Adapter
 *   Token-Ring Network 16/4 Adapter/A (short card)
 *   16/4 ISA-16 Adapter
 *   Auto 16/4 Token-Ring ISA Adapter
 *   Auto 16/4 Token-Ring MC Adapter <-- Not supported currently
 *
 * 
 */

#include "tok.h"
#include "opt_inet.h"

#if NTOK > 0 

#include "bpfilter.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/types.h>
#include <sys/conf.h> 
#include <sys/errno.h>
#include <sys/sockio.h> 
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/syslog.h>

#include <net/if.h>
#include <net/if_arp.h>
#include <net/if_llc.h>
#include <net/if_dl.h>
#include <net/if_mib.h>
#include <net/if_types.h>
#include <net/netisr.h>

#include <net/iso88025.h>

#ifdef INET
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/if_ether.h>
#endif

#ifdef IPX
#include <netipx/ipx.h>
#include <netipx/ipx_if.h>
#endif

#ifdef NS
#include <netns/ns.h>
#include <netns/ns_if.h>
#endif

#if NBPFILTER > 0
#include <net/bpf.h>
#include <net/bpfdesc.h>
#endif

#include <machine/clock.h>
#include <machine/md_var.h>

#include <i386/isa/isa_device.h>
/*#include <i386/isa/icu.h>*/

#include <vm/vm.h>
#include <vm/vm_param.h>
#include <vm/pmap.h>

#include "i386/isa/if_tokreg.h"
#include "i386/isa/if_tokvar.h"

/*
 * Things to be moved into an iso88025 file 
 */

#include <net/route.h>
#include <sys/kernel.h>
#include <net/iso88025.h>

/*
#define ISO88025MTU	1492 
#define senderr(e) { error = (e); goto bad;}
void   iso88025_ifattach __P((struct ifnet *));
int    iso88025_ioctl __P((struct ifnet *, int , caddr_t ));
int    iso88025_output __P((struct ifnet *, struct mbuf *, struct sockaddr *, struct rtentry *));
void   iso88025_input __P((struct ifnet *, struct iso88025_header *, struct mbuf *));
*/

#define TR_DEBUG 1
#define DEBUG_MSG(x) if (TR_DEBUG>0) printf("tok: %s\n", x);

#define TOK_INTR_BRAKE 500
static int tok_intr_count = 0;

#define addrtoisa(x) (caddr_t) ((u_long)0xf0000000 + (u_long)(x))
#define isatoaddr(x) (caddr_t) ((u_long)(x) - (u_long)0xf0000000)
#define AIP_IN(x) (*(caddr_t)(x) & 0xf)
#define MM_OUT(addr, byte)  ((*(caddr_t)(addr)) = (u_char)(byte))
#define MM_OUTW(addr, shrt) ((*(caddr_t)(addr)) = (u_short)htons((u_short)(shrt)))
#define MM_IN(addr) (*(u_char *)(addr))
#define MM_INW(addr)  ((u_short)(ntohs((*(u_short *)(addr)))))
#define DUMP_REG(sc, mesg) printf("tok: isrp e[%x] o[%x] isra e[%x] o[%x] (%s).\n", \
        MM_IN(&sc->aca->rw.isrp_e), MM_IN(&sc->aca->rw.isrp_o), \
        MM_IN(&sc->aca->rw.isra_e), MM_IN(&sc->aca->rw.isra_o), mesg);

/*
 * These are things that should be calculated depending on the
 * card.  But for now we will just make assupmtions. (XXX)
 */
#define NUM_RCV_BUF   16
#define RCV_BUF_LEN   136
#define NUM_DHB       2

static char isa_id[12]={0x50, 0x49, 0x43, 0x4f, 0x36, 0x31, 0x31, 0x30, 0x39, 0x39, 0x30, 0x20};
static char mca_id[12]={0x4d, 0x41, 0x52, 0x53, 0x36, 0x33, 0x58, 0x34, 0x35, 0x31, 0x38, 0x20};

struct tok_softc {
    struct arpcom arpcom;
    u_char   status;
    int      id;
    int      pio;
    caddr_t  mmio;
    caddr_t  sram;
    u_short  irq;
    u_char   type;
    struct   aca *aca;
    struct   cb cb;
    struct   cb init_cb;
    u_char   station_id[2];
    u_char   alive;
    u_char   wire_fault;
    struct callout_handle tok_timeout_ch;
    caddr_t  tok_sleep_event;
    struct mbuf *pkt;
};

static struct tok_softc tok_softc[NTOK];
static u_char num_cards = 0;

static int  tok_probe __P((struct isa_device *));
static int  tok_attach __P((struct isa_device *));
static void tok_init __P((struct tok_softc *));
static void tok_intr __P((int unit));  
static void tok_rx_intr __P((int unit));
static void tok_tx_intr __P((int unit));
static void tok_start __P((struct ifnet *));
static void tok_stop __P((struct tok_softc *));
static int  tok_ioctl __P((struct ifnet *, u_long, caddr_t));
static void tok_watchdog __P((struct ifnet *));
void tok_timeout __P((void *));
static void tok_sleep __P((int));
void tok_open_sap (register int, unsigned char);

void tok_info(int, caddr_t);

struct isa_driver tokdriver = {
    tok_probe,
    tok_attach,
    "tok",
    0   
};  

int
tok_probe(is)
    struct isa_device *is;
{

    u_char   setup_read_1 = 0, setup_read_2 = 0,
             isa_ch_id, mca_ch_id, adapter_type, i, j, k;
    u_short  irq = 0;
    caddr_t  mmio = 0L;
    int      pio;
    u_char   chan_id[12];

    DEBUG_MSG("tok_probe called.");

    /* Only two cards maximum, make sure we dont exceed that */
    if (num_cards == 2) 
        return(0);

    /* Check for valid port address */ 
    if ((is->id_iobase != PIO_PRI) && (is->id_iobase != PIO_ALT)) 
         return(0);
    else 
         pio = is->id_iobase;

    /* Make sure we dont probe the same card twice */
    if (num_cards > 0) {
        for (i = 0; i < num_cards; i++) {
            if (tok_softc[i].pio == pio) 
                return(0);
        }
    }

    /* Get setup info from PIO port then check for bogus contents. */ 
    setup_read_1 = inb(pio);
    setup_read_2 = inb(pio + 2);
    if ((setup_read_1 == 0xff) && (setup_read_2 == 0xff)) 
        return(0); 
       
    /* Compute mmio base address for card and check to see if it is in range. */
    mmio = addrtoisa(((setup_read_2 & 0x1) << 19) | ((setup_read_1 & 0xfc) << 11)); 
    if ((isatoaddr(mmio) < (caddr_t)0xc8000) || ((isatoaddr(mmio) > (caddr_t)0xdffff)))
        return(0);
 
    /* Get adapter type (F,E,D,C are valid for shared ram) and make sure it's ours */
    adapter_type = AIP_IN(mmio+0x1fa0);
    if (adapter_type < 0xc)
        return(0);
     
    /* Determine IRQ */
    switch(setup_read_1 & 0x3) {
        case IRQ2: 
             irq = 9;
             break;
        case IRQ3:
             irq = 3;
             break;
        case IRQ6: /* Auto cards use IRQ10 with same mask */
             if (adapter_type != 0xc)
                 irq = 6;
             else
                 irq = 10;
             break;
         case IRQ7: /* Auto cards usr IRQ11 with same mask */
             if (adapter_type != 0xc) 
                 irq = 7;
             else 
                 irq = 11;
             break;
         default:
             break;
     }
   
    /* Check for valid channel identifier in AIP, ISA or MicroChannell (MCA) either
     * will work but we cant attach an MCA adapter.
     */
    isa_ch_id = 1; mca_ch_id = 1;
    for (j = 0, k = 0; j < 48; j+=4) {
        chan_id[k]  = (AIP_IN(mmio+CH_ID_OFFSET+j) << 4) | (AIP_IN(mmio+CH_ID_OFFSET+j+2));
        if ((isa_ch_id) && (chan_id[k] != isa_id[k])) isa_ch_id = 0;
        if ((mca_ch_id) && (chan_id[k] != mca_id[k])) mca_ch_id = 0;
        if (!isa_ch_id && !mca_ch_id)
            break;
        k++;
    }
    if (!isa_ch_id && !mca_ch_id)
        return(0);
 
    /* If you made it here you've got a token ring card. YEAH!
     * At this point the values found need to be punched back out to the
     * isa_device structure, so that the kernel knows where everybody is and
     * what irq's they are on.
     * irq, port -> iobase, mmio -> maddr, (tok_intr) -> intr
     * and oddly enough we now know all of these :)  Otherwise we have to
     * go through all this again, or hide it off in memory for attach. 
     */

    printf("tok%d: IBM Token-Ring Adater type %x found: addr %p irq %d\n", \
            is->id_unit, adapter_type, isatoaddr(mmio), irq);

    tok_softc[is->id_unit].status = UNKNOWN;
    tok_softc[is->id_unit].id = is->id_id;
    tok_softc[is->id_unit].pio = pio;
    tok_softc[is->id_unit].mmio = mmio;
    tok_softc[is->id_unit].irq = irq;
    tok_softc[is->id_unit].type = adapter_type;
    tok_softc[is->id_unit].aca = (struct aca *)(mmio + ACA_OFFSET);

    is->id_irq = (1 << irq ); 
    is->id_intr = (inthand2_t *)tok_intr;

    num_cards++;

    return(1);
}

int
tok_attach(is)
    struct isa_device *is;
{   
    int i = 0, j = 0, init = 0;
    struct tok_softc *sc = &tok_softc[is->id_unit];
    struct ifnet *ifp = &sc->arpcom.ac_if;
    struct init_resp *srb; 
    u_char temp[6]; 
   
    DEBUG_MSG("tok_attach called.");

    sc->wire_fault = 0x00; /* Reset wire fault */

    printf("tok%d: card io_base %p  address %p  irq %d\n",\
        is->id_unit, is->id_iobase, is->id_maddr, ffs(is->id_irq)-1);

    outb(is->id_iobase + 0x1, 1);     /* Adapter reset latch */

    /* Set SRPR bits 14 and 15 to 0xC0 for shared ram paging */   

    DELAY(50000);

    outb(is->id_iobase + 0x2, 1);     /* Adapter reset release */

    tok_softc[is->id_unit].status = RESET;

    MM_OUT(&sc->aca->set.isrp_e, INT_ENABLE);
    for (i = 0; i < 6; i++) {
        DELAY(500000); 
        if ((sc->aca->rw.isrp_o & SRB_RESP_INT) == SRB_RESP_INT) {
            init = 1; 
            break;
        }
    }

    if (init == 0) {
        printf("tok%d: adapter failed to initialize.\n", is->id_unit);
        return(0);
    }

    tok_info(is->id_unit, sc->mmio); /* Print AIP setup information */

    for (i = 0, j = 0; i < 24; i+=4, j++)
        sc->arpcom.ac_enaddr[j]  = (AIP_IN(sc->mmio+AIP_OFFSET+i) << 4) | (AIP_IN(sc->mmio+AIP_OFFSET+i+2));

    /* Set RRR if necessary */
    if (sc->type > 0xc) {
        MM_OUT(&sc->aca->rw.rrr_e, (u_char)((u_int)(is->id_maddr) >> 12));
        sc->sram = is->id_maddr;
    } else {
        sc->sram = addrtoisa((inb(is->id_iobase +0x2) & 0xfe) << 12);
    }

    /* Compute location of first srb */
    sc->cb.srb = (union sr_srb *)(sc->sram + ((sc->aca->rw.wrbr_e << 8) | (sc->aca->rw.wrbr_o)));
    sc->init_cb.srb = sc->cb.srb;
    srb = (struct init_resp *)(sc->cb.srb);

    printf("tok%d: adapter shared ram address %p\n", is->id_unit, sc->sram);
    printf("tok%d: RRR contents: %x\n", is->id_unit, sc->aca->rw.rrr);
    printf("tok%d: Initial srb location %p.\n", is->id_unit, sc->cb.srb);
    printf("tok%d: srb command [0x%x]\n", is->id_unit, srb->cmd);
    printf("tok%d: init status [0x%x 0x%x]\n", is->id_unit, srb->status, srb->status_2);
    printf("tok%d: bring up code [0x%x]\n", is->id_unit, srb->bring_up_code);
    for (i = 0; i < 6; i++)
        temp[i] = MM_IN(sc->sram + ntohs(srb->encoded_addr) + i);
    printf("tok%d: MAC - %6D\n", is->id_unit, temp, ":");


    /* Initialize the ifnet structure. */
    ifp->if_softc = sc;
    ifp->if_unit = is->id_unit;
    ifp->if_name = "tok";
    ifp->if_output = iso88025_output;
    ifp->if_init = (if_init_f_t*)tok_init;
    ifp->if_start = tok_start;
    ifp->if_ioctl = tok_ioctl;
    ifp->if_flags = IFF_BROADCAST | IFF_MULTICAST | IFF_SIMPLEX;
    ifp->if_watchdog = tok_watchdog;
    ifp->if_snd.ifq_maxlen = IFQ_MAXLEN;

    /* Initialize callout handle for interface */
    /*struct callout_handle sc->tok_timeout_ch = CALLOUT_HANDLE_INITIALIZER(&sc->tok_timeout_ch);*/
    callout_handle_init(&sc->tok_timeout_ch);

    /* Attach the interface. */
    if_attach(ifp);
    iso88025_ifattach(ifp);

#if NBPFILTER > 0
    bpfattach(ifp, DLT_IEEE802, sizeof(struct iso88025_header));
#endif    
    
    /* Free the init srb */
    MM_OUT(&sc->aca->reset.isrp_o, ~(SRB_RESP_INT));
    MM_OUT(&sc->aca->reset.isra_o, ~(CMD_IN_SRB));
    MM_OUT(&sc->aca->reset.isrp_e, ~(INT_ENABLE));

    sc->alive = 1;

    return(1);
}

void
tok_info(unit, mmio)
    int  unit;
    caddr_t mmio;
{
    printf("tok%d: Adapter information:\n", unit);
    printf("      type: %x\n", AIP_IN(mmio+0x1fa0));
    printf("      data rate: %x\n", AIP_IN(mmio+0x1fa2));
    printf("      early token release: %x\n", AIP_IN(mmio+0x1fa4));
    printf("      available shared ram: %x\n", AIP_IN(mmio+0x1fa6));
    printf("      shared ram paging: %x\n", AIP_IN(mmio+0x1fa8));
    printf("      4 Mbps data holding buffer: %x\n", AIP_IN(mmio+0x1faa));
    printf("      16 Mbps data holding buffer: %x\n", AIP_IN(mmio+0x1fac));
    printf("      STP/UTP support: %x\n", AIP_IN(mmio+0x1fb6));
    printf("      STP/UTP selected: %x\n", AIP_IN(mmio+0x1fb8));
    printf("      interrupts supported: %x\n", AIP_IN(mmio+0x1fba));
}

void
tok_timeout(unit_p)
    void *unit_p;
{
    int unit = (int)unit_p;
    register struct tok_softc *sc = &tok_softc[unit];

    printf("tok%d: adapter operation timeout.\n", unit);
    DUMP_REG(sc, "tok_timeout");
    wakeup(&sc->tok_sleep_event);

    return;
}

void
tok_sleep(unit)
    register int unit;
{
    register struct tok_softc *sc = &tok_softc[unit];

    timeout((timeout_t *)tok_timeout, (void *)unit, 30*hz);
    tsleep(&sc->tok_sleep_event, 1, "tokint", 30*hz);
}

void
tok_intr(unit)
    int unit;
{
    register struct tok_softc *sc = &tok_softc[unit];
    register struct ifnet *ifp = &sc->arpcom.ac_if;
    register union  sr_srb *srb;
    register union  sr_asb *asb;
    register union  sr_arb *arb;
    register struct sr_ssb *ssb;


    caddr_t  check_code;

    /* Adapter status register */
    register u_char status;

    /* Command from status or request block */
    register u_char command;

    u_char i;
    int    rc = 1; 
    int    close_signal = 0;

    if (TOK_INTR_BRAKE) tok_intr_count++;

    /* Hold interrupts while we work */
    MM_OUT(&tok_softc[unit].aca->reset.isrp_e, ~(INT_ENABLE)); 

    outb(sc->pio + 0x3, 1);

    if (tok_softc[unit].status == RESET) {
        tok_softc[unit].status = READY;
    }

    printf("tok%d: Interrupt received.\n", unit);
    DUMP_REG(sc, "tok_intr");
  
    if (tok_softc[unit].status == UNKNOWN) {
        printf("tok%d: stray interrupt from uninitialized adapter.\n", unit);
        return;
    }

    status = MM_IN(&sc->aca->rw.isrp_o);

    if (TOK_INTR_BRAKE) {
        if (tok_intr_count > TOK_INTR_BRAKE) {
            printf("tok: TOK_INTR_BRAKE exceeded\n");
            goto brake;
        }
    }
    
    /* Look for adapter check */
    if (status & ADAPT_CHK_INT) {
        check_code = MM_INW(&sc->aca->rw.wwcr) + sc->sram;
        printf("tok%d: adapter check (%x).\n", unit, MM_INW(&check_code));
        /* Clear adapter check interrrupt */
        MM_OUT(&sc->aca->reset.isrp_o, ~(status));
        MM_OUT(&sc->aca->reset.isrp_o, ~(ADAPT_CHK_INT));
        MM_OUT(&sc->aca->set.isrp_e, INT_ENABLE);

        rc = 0;  /* Claim interrupt. */
        goto done;
    }

    /* Check for adapter errors or access violations */
    else if ((i = MM_IN(&sc->aca->rw.isrp_e)) & (TVR_INT + ERR_INT + ACCESS_INT)) {
        printf("tok%d: adapter error (isrp_e = 0 %x).\n", unit, i);

        /* Clear error interrupt bits */
        MM_OUT(&sc->aca->reset.isrp_e, ~(TVR_INT + ERR_INT + ACCESS_INT));
        MM_OUT(&sc->aca->set.isrp_e, INT_ENABLE);
    
        rc = 0;  /* Claim interrupt */
    } 

    /* Check for normal operational interrupts */
    else if (status & (SRB_RESP_INT + ASB_FREE_INT + ARB_REQ_INT + SSB_RESP_INT)) {

        /* Adapter response to SRB command */
        if (status & SRB_RESP_INT) {
            srb = (union sr_srb *)sc->cb.srb;
 
            command = MM_IN(&srb->xmit.cmd);
            switch(command) {
            
                case XMIT_DIR_FRAME:                        /* Response to xmit */
                case XMIT_UI_FRAME:
                    printf("tok%d: Transmit SRB response.\n", unit);
                    if ((i = MM_IN(&srb->xmit.retcode)) != 0xff) {
                        printf("tok%d: SRB error on transmit request (rc=%x).\n", unit, i);
                        ifp->if_flags &= ~IFF_OACTIVE;
                    } else  {
                        ifp->if_opackets++;
                        ifp->if_flags &= ~IFF_OACTIVE;
                    }
                    break;
                case DIR_OPEN_ADAPTER:                      /* Open adapter response */
                    printf("tok%d: Open adapter SRB response.\n", unit);
                    if ((i = MM_IN(&srb->open_resp.retcode)) == 0) {
                        ifp->if_flags |= (IFF_UP | IFF_RUNNING);
                        ifp->if_flags &= ~IFF_OACTIVE;

                        
                        printf("tok%d:  SRB [0x%x] (rc=0x%x codes[0x%x 0x%x]).\n", unit, command, \
                                srb->open_resp.retcode, srb->open_resp.open_error_code[0], \
                                srb->open_resp.open_error_code[1]);
                                
                        /* Save new ACA control buffer addresses */
                        sc->cb.srb = (union sr_srb *) (MM_INW(&srb->open_resp.srb_address) + sc->sram);
                        sc->cb.ssb = (struct sr_ssb *)(MM_INW(&srb->open_resp.ssb_address) + sc->sram);
                        sc->cb.arb = (union sr_arb *) (MM_INW(&srb->open_resp.arb_address) + sc->sram);
                        sc->cb.asb = (union sr_asb *) (MM_INW(&srb->open_resp.asb_address) + sc->sram);

                        /*MM_OUT(&sc->aca->set.isrp_e, INT_ENABLE); */

                        wakeup(&sc->tok_sleep_event);

                        /* Reset wire fault */
                        sc->wire_fault = 0x00;
                     } else {
                         printf("tok%d: Open adapter error (rc = %x).\n", unit, i);
                         ifp->if_flags &= ~(IFF_UP | IFF_RUNNING);
                         sc->tok_timeout_ch = timeout((timeout_t *)tok_init, &ifp->if_unit, 30*hz);
                     }
                     break;
                case DIR_CLOSE_ADAPTER:                        /* Close adapter response */
                    printf("tok%d: Close adapter SRB response.\n", unit);
                    if ((i = MM_IN(&srb->close.retcode)) != 0)
                        printf("tok%d: Close adapter error (rc = %x).\n", unit, i);
                    else if (i == 0) {
                        close_signal = 1;
                        ifp->if_flags &= ~(IFF_UP | IFF_RUNNING);
                        MM_OUT(&sc->aca->reset.isrp_e, ~(INT_ENABLE));
                        sc->cb.srb = (union sr_srb *)(sc->sram + ((sc->aca->rw.wrbr_e << 8) | (sc->aca->rw.wrbr_o)));
                        /* Disable Interrupts (XXX) */
                        MM_OUT(&sc->aca->reset.isrp_o, ~(status));
                        MM_OUT(&sc->aca->reset.isrp_o, ~(SRB_RESP_INT));
                        MM_OUT(&sc->aca->reset.isrp_e, ~(INT_ENABLE));
                        return;
                    }
                    break;

                case DIR_INTERRUPT:                             /* Driver interrupt response */
                    printf("tok%d: Interrupt adapter SRB response.\n", unit);
                    if ((i = MM_IN(&srb->interrupt.retcode)) != 0)
                        printf("tok%d: Forced adapter interrupt failed (rc = %x).\n", unit, i);
                    break;

                case DIR_MODIFY_OPEN_PARMS:                     /* Modify open params response */
                    printf("tok%d: Modify open parameters SRB response.\n", unit);
                    if ((i = MM_IN(&srb->mod_params.retcode)) != 0) 
                        printf("tok%d: Forced adapter interrupt failed (rc = %x).\n", unit, i);
                    break; 

                case DIR_SET_GROUP_ADDRESS:                     /* Set group address response */
                case DIR_SET_FUNCTIONAL_ADDRESS:                /* Set functional address response */
                    printf("tok%d: Set functional/group address SRB response.\n", unit);
                    if ((i = MM_IN(&srb->set_addr.retcode)) != 0)
                        printf("tok%d: Set group or functional address failed (rc = %x).\n", unit, i);
                    break;

                case DLC_OPEN_SAP:                              /* Open sap response */
                    printf("tok%d: Open DLC SAP SRB response.\n", unit);
                    /* Mach checks for extended sap (no information?) */
                    if ((i = MM_IN(&srb->open_sap.retcode)) != 0) {
                        printf("tok%d: Ring opened failed.\n", unit);
                    } else {
                        sc->station_id[0] = MM_IN(&srb->open_sap.station_id[0]);
                        sc->station_id[1] = MM_IN(&srb->open_sap.station_id[1]);
                        ifp->if_flags |= (IFF_UP | IFF_RUNNING);
                        printf("tok%d: SAP station [%x%x] [%x]\n", unit, sc->station_id[0], sc->station_id[1], MM_INW(&srb->open_sap.station_id));
                    }
                        wakeup(&sc->tok_sleep_event);
                    break;

                case DLC_CLOSE_SAP:                             /* Close sap response */
                    printf("tok%d: Close DLC SAP SRB response.\n", unit);
                    break;

                case DIR_READ_LOG:                              /* Read log response */
                    printf("tok%d: Read log SRB response.\n", unit);
                    if ((i = MM_IN(&srb->read_log.retcode)) != 0)
                        printf("tok%d: Read log failed (rc = %x).\n", unit, i);
                    /* Free SRB */
                    break;

                default:
                    printf("tok%d: Bad command in SRB (0x%x).\n", unit, command);
                    break;
            } 
            /* Clear SRB response interrupt bit */
            printf("tok%d: Clearing interrupt.\n", unit);
            MM_OUT(&sc->aca->reset.isrp_o, ~(SRB_RESP_INT));

            /* Start output on interface */
            /*if (command == DLC_OPEN_SAP && i == 0) {*/
                /* Free SRB */
                /* Start output on interface */
                /*tok_start(ifp);*/
            /*}*/
        }     

        /* Adapter request to Free ASB */
        if (status & ASB_FREE_INT) {
            asb = (union sr_asb *)sc->cb.asb;

            command = MM_IN(&asb->rec_data.cmd);
            switch (command) {

                case RECEIVED_DATA:                            /* Receive Data */
                    if ((i = MM_IN(&asb->rec_data.retcode)) != 0xff)
                        printf("tok%d: Receive data failed (rc = %x).\n", unit, i);
                    break;

                case XMIT_DIR_FRAME:                           /* Transmit */
                    if ((i = MM_IN(&asb->data_req.retcode)) != 0xff) {
                        printf("tok%d: Transmit data request failed (rc = %x).\n", unit, i);
                        MM_OUT(&sc->aca->set.isra_o, ARB_FREE);
                        MM_OUT(&sc->aca->set.isra_o, ASB_FREE);
                    }
                    break;

                case XMIT_UI_FRAME:                            /* Transmit UI */
                     if((i = MM_IN(&asb->data_req.retcode)) != 0xff)
                         printf("tok%d: Transmit UI request failed (rc = %x).\n", unit, i);
                     break;

                default:
                     printf("tok%d: Bad command in ASB (0x%x).\n", unit, command);
                     break;
            }

            /* Clear ASB reqest interrupt bit */
            MM_OUT(&sc->aca->reset.isrp_o, ~(ASB_FREE_INT));
        }

        /* Adapter request for driver to process command */
        if (status & ARB_REQ_INT) {
            arb = (union sr_arb *)sc->cb.arb;

            command = MM_IN(&arb->dlc_status.cmd);
            switch(command) {
                case DLC_STATUS:                                /* DLC Status Change */
                    printf("tok%d: DLC status change (status = 0x%x).\n", unit, MM_IN(&arb->dlc_status.status));
                    break;

                case RECEIVED_DATA:                             /* Data waiting for driver */
                    tok_rx_intr(unit);
                    break;

                case RING_STATUS_CHANGE:                        /* Ring state changed */
                    if ((i == MM_INW(&arb->ring_status_change.status)) & (SIGNAL_LOSS + WIRE_FAULT)) {
                        printf("tok%d: Signal loss or wire fault on node (status 0x%d).\n", unit, i);
                        ifp->if_flags &= ~(IFF_RUNNING | IFF_UP);
                        sc->wire_fault = 0x01;
                        sc->tok_timeout_ch = timeout((timeout_t *)tok_init, &ifp->if_unit, 30*hz);
                    } else {
                        if (i & ~SOFT_ERR)
                            printf("tok%d: Ring status change (status 0x%x).\n", unit, i);
                    }
                    if (i & COUNTER_OVERFLOW) {
                        MM_OUT(&sc->cb.srb->read_log, DIR_READ_LOG);
                        MM_OUT(&sc->aca->set.isrp_e, INT_ENABLE);
                        MM_OUT(&sc->aca->set.isra_o, CMD_IN_SRB);
                        printf("tok%d: Reading adapter error log (log overflow).\n", unit);
                    }
                    break;
                
                case TRANSMIT_DATA_REQUEST:
                    printf("tok%d: Transmit data request ARB recieved.\n", unit);
                    tok_tx_intr(unit);
                    break;
                
                default:
                    printf("tok%d: Bad command in ARB (0x%x).\n", unit, command);
                    break;
                }
            /* Clear ARB interrupt and free ARB */
            MM_OUT(&sc->aca->reset.isrp_o, ~(ARB_REQ_INT));
            MM_OUT(&sc->aca->set.isra_o, ARB_FREE);
        }

        if (status & SSB_RESP_INT) {
            ssb = (struct sr_ssb *)sc->cb.ssb;

            command = MM_IN(&ssb->cmd);
            switch(command) {
                case XMIT_DIR_FRAME:
                case XMIT_UI_FRAME:
                    printf("tok%d: Transmit_ui/dir_frame SSB received.\n", unit);
                    if ((i == MM_IN(&ssb->retcode)) != 0) {
                        printf("tok%d: Error on transmit_dir_frame (rc=%d).\n", unit, i);
                        ifp->if_oerrors++;
                        ifp->if_flags &= ~IFF_OACTIVE;
                    } else {
                        ifp->if_opackets++;
                        ifp->if_flags &= ~IFF_OACTIVE;
                    }
                    break;
                case XMIT_XID_CMD:
                    i = MM_IN(&ssb->retcode);
                    printf("tok%d: Transmit_XID_command received (rc=%x).\n", unit, i);
                    
                default:
                    printf("tok%d: Bad command in SSB (0x%x).\n", unit, command);
                    break;
            }

            MM_OUT(&sc->aca->reset.isrp_o, ~(SSB_RESP_INT));
            MM_OUT(&sc->aca->set.isra_o, SSB_FREE);
        }
    
    }

done:
    /* Enable interrupts */
    /*MM_OUT(&sc->aca->reset.isrp_o, ~(status));*/
    MM_OUT(&sc->aca->set.isrp_e, INT_ENABLE);

    return;

brake:
    /* Enable interrupts */
    MM_OUT(&sc->aca->reset.isrp_o, ~(status));
    MM_OUT(&sc->aca->set.isrp_e, INT_ENABLE);

    return;

}

static void
tok_rx_intr(unit)
    int unit;
{
    register struct tok_softc *sc = &tok_softc[unit];
    register union  sr_arb *arb = sc->cb.arb;
    register union  sr_asb *asb = sc->cb.asb;
    struct ifnet *ifp = &sc->arpcom.ac_if;
    struct rec_buf *buffer;

    u_char rc = 0x00;

    struct iso88025_header *th;
    struct mbuf *m;

    int frame_len, dlc_hdr_len, lan_hdr_len, bytes_copied, len;

    printf("tok%d: Receive interrupt called.\n", unit);

    frame_len = MM_INW(&arb->rec_data.frame_length);
    dlc_hdr_len = MM_IN(&arb->rec_data.dlc_hdr_length);
    lan_hdr_len = MM_IN(&arb->rec_data.lan_hdr_length);

    printf("tok%d: Packet received data:", unit);
    printf("  cmd[%x] station_id[%x%x]", 
            MM_IN(&arb->rec_data.cmd), MM_IN(&arb->rec_data.station_id[0]), MM_IN(&arb->rec_data.station_id[1]));
    printf(" bptr[%x] lhdr[%x] dhdr[%x] len[%x] ncb[%x]\n", 
            MM_INW(&arb->rec_data.receive_buffer), lan_hdr_len, dlc_hdr_len, frame_len, MM_IN(&arb->rec_data.ncb_type));

    /* Allocate a header mbuf */
    MGETHDR(m, M_DONTWAIT, MT_DATA);
    if (m == NULL) {
        ifp->if_ierrors++;
        rc = 0x20;
        goto out;
    }
    m->m_pkthdr.rcvif = &sc->arpcom.ac_if;
    m->m_pkthdr.len = m->m_len = frame_len;

    /* 
     * See if we can get the packet in one buffer, if not then allocate
     * a cluster for the packet.  The +2 is for NFS alignment.
     */
    if (frame_len + 2 > MHLEN) {
        /* Get an mbuf cluster */
        MCLGET(m, M_DONTWAIT);
        /* Make sure we got our cluster or free the header and give up. */
        if ((m->m_flags & M_EXT) == 0) {
            m_freem(m);
            ifp->if_ierrors++;
            rc = 0x20;
            goto out;
        }
    }
    ifp->if_ipackets++;

    /* Move the data pointer to fixup the NFS word alignment. */
    m->m_data += 2; 
    th = mtod(m, struct iso88025_header *);

    /* Pull the packet off the interface and into the mbuf */
    bytes_copied = 0;
    buffer = (struct rec_buf *)(sc->sram + MM_INW(&arb->rec_data.receive_buffer));
    while (bytes_copied < frame_len) {
        len = min(MM_INW(&buffer->buffer_length), frame_len - bytes_copied);
        bcopy(&buffer->frame_data, mtod(m, char *) + bytes_copied, len);
        bytes_copied += len;
        buffer = (struct rec_buf *)(sc->sram + MM_INW(&buffer->next_buffer) - 2);
    }

#if NBPFILTER > 0
    if (ifp->if_bpf)
        bpf_mtap(ifp, m);
#endif

    /* Adjust the mbuf pointer past the MAC header */
    m->m_pkthdr.len = m->m_len = frame_len - lan_hdr_len;
    m->m_data += lan_hdr_len;

    iso88025_input(&sc->arpcom.ac_if, th, m);

out:
    MM_OUT(&asb->rec_data.cmd, RECEIVED_DATA);
    MM_OUT(&asb->rec_data.retcode, rc);
    MM_OUT(&asb->rec_data.station_id[0], MM_IN(&arb->rec_data.station_id[0]));
    MM_OUT(&asb->rec_data.station_id[1], MM_IN(&arb->rec_data.station_id[1]));
    MM_OUT(&asb->rec_data.receive_buffer, MM_IN(&arb->rec_data.receive_buffer));
    MM_OUT(&asb->rec_data.receive_buffer[1], MM_IN(&arb->rec_data.receive_buffer[1]));
    MM_OUT(&sc->aca->set.isra_o, RESP_IN_ASB);

    return;
}

static void
tok_tx_intr(unit)
    int unit;
{
    struct tok_softc *sc = &tok_softc[unit];
    struct ifnet *ifp = &sc->arpcom.ac_if;
    union  sr_arb *arb = (union sr_arb *)sc->cb.arb;
    union  sr_asb *asb = (union sr_asb *)sc->cb.asb;
    u_char *dhb;
    struct mbuf *m, *m0;
    int    length, s, i;

    s = splimp();

    printf("tok%d: Transmit interrupt called.\n", unit);
    printf("tok: sanity check %s\n", ifp->if_name);

    IF_DEQUEUE(&ifp->if_snd, m);
    if(m == 0) {
        printf("tok%d: tok_tx_int NULL packet?\n", unit);
        ifp->if_flags &= ~IFF_OACTIVE;
        splx(s);
        return;
    }       
    m0 = m;
        
    /* Create transmit ASB */
    printf("tok%d: creating ARB station_id[%x%x]].\n", unit, sc->station_id[0], sc->station_id[1]);
    MM_OUT(&asb->data_req.cmd, XMIT_UI_FRAME);
    MM_OUT(&asb->data_req.cmd_correlate, MM_IN(&arb->xmit_request.cmd_correlate));
    MM_OUT(&asb->data_req.station_id[0], sc->station_id[0]);
    MM_OUT(&asb->data_req.station_id[1], sc->station_id[1]);
    MM_OUT(&asb->data_req.retcode, 0);
    MM_OUT(&asb->data_req.rsap_value, 0xaa);
    MM_OUT(&asb->data_req.header_length, ISO88025_HDR_LEN);

    /* Get dhb address and start to fill it out */
    dhb = (u_char *)(sc->sram + MM_INW(&arb->xmit_request.dhb_address));

    /* Copy mbufs into adapter dhb */
    printf("tok%d: Copying mbufs.\n", unit);
    for (length = 0; m != 0; m = m->m_next) {
        bcopy(mtod(m, caddr_t), dhb + length, m->m_len);
        length += m->m_len;
    }
    
    /* Set the size of the frame in the ASB */
    /*MM_OUTW(&asb->data_req.frame_length, htons(length));*/
    MM_OUT(&asb->data_req.frame_length[0], (length & 0x00FF));
    MM_OUT(&asb->data_req.frame_length[1], (length & 0xFF00) >> 8);

    /* Interrrupt the adapter */
    printf("tok%d: interrupting adapter.\n", unit);
    MM_OUT(&sc->aca->set.isra_o, RESP_IN_ASB);

    /*
     * Tap off here if there is a bpf listener.
     */
#if NBPFILTER > 0
    if (ifp->if_bpf) {
        printf("tok%d: copying packet to adapter.\n", unit);
        bpf_mtap(ifp, m0);
    }
#endif
    
    m_freem(m);

    ifp->if_flags &= ~IFF_OACTIVE;

    printf("tok: reset OACTIVE.\n");

    (void) splx(s);

}

void
tok_init(sc)
    struct tok_softc *sc;
{
    struct ifnet *ifp = &sc->arpcom.ac_if;
    union sr_srb *srb = sc->cb.srb;
    int s = 0, i = 0;

    printf("tok: tok_init called (srb %p).\n", srb);

    /* address not known */
    if (TAILQ_EMPTY(&ifp->if_addrhead)) /* unlikely? XXX */
        return;

    if ((ifp->if_flags & IFF_RUNNING) == 0) {
        s = splimp();

        /* 
         * Set 'running' flag, and clear output active flag.
         */
        /*ifp->if_flags |= IFF_RUNNING;
        ifp->if_flags &= ~IFF_OACTIVE;*/

        /* Clear out the srb so we can skip things that get defaulted */
        for (i = 0; i < sizeof(union sr_srb); i++)
            MM_OUT(((char *)srb + i), 0);
        /* Set open parameters */
        MM_OUT(&srb->open_cmd.cmd, DIR_OPEN_ADAPTER); 
        MM_OUTW(&srb->open_cmd.open_options, OPEN_PASS_BCON_FRAMES);
        MM_OUTW(&srb->open_cmd.num_rcv_buf, NUM_RCV_BUF);
        MM_OUTW(&srb->open_cmd.rcv_buf_len, RCV_BUF_LEN);
        MM_OUTW(&srb->open_cmd.num_dhb, NUM_DHB);
        MM_OUT(&srb->open_cmd.dlc_max_sap, 2);
        MM_OUT(&srb->open_cmd.dlc_max_sta, 8);

        /* Interrupt the adapter */
        outb(sc->pio + 0x3, 1);
        MM_OUT(&sc->aca->set.isrp_e, INT_ENABLE);
        MM_OUT(&sc->aca->set.isra_o, CMD_IN_SRB);

        tsleep(&sc->tok_sleep_event, 1, "tokopn", 30*hz);

        tok_open_sap(ifp->if_unit, 0xaa);

        tsleep(&sc->tok_sleep_event, 1, "toksap", 30*hz);

        (void) splx(s);
    }
    return;
}
void 
tok_stop(sc)
    struct tok_softc *sc;
{
    register union sr_srb *srb = sc->cb.srb;
    register struct ifnet *ifp = &sc->arpcom.ac_if;

    tok_intr_count = 0;

    printf("tok: tok_stop called (srb %p).\n", srb);

    if (ifp->if_flags & IFF_RUNNING) {

        MM_OUT(&sc->aca->reset.isrp_e, ~(S_INT_MASK));

        /* Issue close command */
        MM_OUT(&srb->close.cmd, DIR_CLOSE_ADAPTER);
        MM_OUT(&sc->aca->set.isrp_e, INT_ENABLE);
        MM_OUT(&sc->aca->set.isra_o, CMD_IN_SRB);
    }

    return;
}

void 
tok_start(ifp)
    struct ifnet *ifp;
{
    struct tok_softc *sc = ifp->if_softc;
    union sr_srb *srb = sc->cb.srb;
    /*struct mbuf *m;*/
    int s;

    s = splnet();

    if (ifp->if_flags & IFF_OACTIVE) {
       splx(s);
       return;
    } 

    printf("tok: tok_start called.\n");

    MM_OUT(&srb->xmit.cmd, XMIT_UI_FRAME);
    /*MM_OUT(&srb->xmit.cmd, 0x00);*/
    MM_OUT(&srb->xmit.retcode, 0x00);
    MM_OUT(&srb->xmit.reserved1, 0x00);
    MM_OUT(&srb->xmit.station_id[0], sc->station_id[0]); /*sc->station_id[0]);*/
    MM_OUT(&srb->xmit.station_id[1], sc->station_id[1]); /*sc->station_id[1]);*/

    MM_OUT(&sc->aca->set.isra_o, (CMD_IN_SRB | SRB_FREE));

    splx(s);

    return;

}

static int 
tok_ioctl(ifp, cmd, data)
    register struct ifnet *ifp;
    u_long cmd;
    caddr_t data;
{
    struct tok_softc *sc = ifp->if_softc;
    int s, error = 0;

    s = splimp();

    printf("tok: tok_ioctl called.\n");

    switch (cmd) {

        case SIOCSIFADDR:
        case SIOCGIFADDR:
            printf("tok: SIOCSIFADDR/SIOCGIFADDR\n");
            iso88025_ioctl(ifp, cmd, data);
            tok_init(ifp->if_softc);
            break;

        case SIOCSIFFLAGS:
            if ((ifp->if_flags & IFF_UP) == 0 &&
                (ifp->if_flags & IFF_RUNNING) != 0) {
                    /*
                     * If interface is marked down and it is running, then
                     * stop it.
                     */
                    printf("tok: SIOCSIFFLAGS bringing interface down.\n");
                    tok_stop(sc);
                    ifp->if_flags &= ~IFF_RUNNING;
            } else if ((ifp->if_flags & IFF_UP) != 0 &&
                       (ifp->if_flags & IFF_RUNNING) == 0) {
                    /*
                     * If interface is marked up and it is stopped, then
                     * start it.
                     */
                    printf("tok: SIOCSIFFLAGS bringing interface up.\n");
                    tok_init(ifp->if_softc);
            } else {
            /*
             * XXX deal with flags changes:
             * IFF_MULTICAST, IFF_PROMISC,
             * IFF_LINK0, IFF_LINK1,
             */
            }
            break;

        case SIOCSIFMTU:
            printf("tok: SIOCSIFMTU\n");
            iso88025_ioctl(ifp, cmd, data);
            tok_init(ifp->if_softc);
            break;

        default:
            error = EINVAL;
            break;
    }

    splx(s);

    return(error);
}

void 
tok_open_sap(unit, sap)
    register int unit;
    unsigned char sap;
{
    register struct tok_softc *sc = &tok_softc[unit];
    register union  sr_srb *srb = sc->cb.srb;

    printf ("tok%d: Opening sap type %x (srb %p).\n", unit, sap, srb);

    MM_OUT(&sc->aca->reset.isrp_o, ~(SRB_RESP_INT));

    MM_OUT(&srb->open_sap.cmd, DLC_OPEN_SAP);
    MM_OUT(&srb->open_sap.retcode, 0x00);
    MM_OUT(&srb->open_sap.station_id[0], 0x00);
    MM_OUT(&srb->open_sap.station_id[1], 0x00);
    MM_OUT(&srb->open_sap.timer_t1, 0x00);
    MM_OUT(&srb->open_sap.timer_t2, 0x00);
    MM_OUT(&srb->open_sap.timer_ti, 0x00);
    MM_OUT(&srb->open_sap.maxout, 0x00);
    MM_OUT(&srb->open_sap.maxin, 0x00);
    MM_OUT(&srb->open_sap.maxout_incr, 0x00);
    MM_OUT(&srb->open_sap.max_retry_count, 0x00);
    MM_OUT(&srb->open_sap.gsap_max_mem, 0x00);
    MM_OUTW(&srb->open_sap.max_i_field, 0x0088);
    MM_OUT(&srb->open_sap.sap_value, sap);
    MM_OUT(&srb->open_sap.sap_options, 0x24);
    MM_OUT(&srb->open_sap.station_count, 0x01);
    MM_OUT(&srb->open_sap.sap_gsap_mem, 0x00);

    MM_OUT(&sc->aca->set.isra_o, CMD_IN_SRB);

    return;
}
        
void tok_watchdog(struct ifnet *ifp)
{
    printf("tok: tok_watchdog called.\n");
}

#endif
