/*  
 * Copyright (c) 1998, Larry Lile
 * All rights reserved.
 *
 * For latest sources and information on this driver, please
 * go to http://anarchy.stdio.com.
 *
 * Questions, comments or suggestions should be directed to
 * Larry Lile <lile@stdio.com>.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: if_tokvar.h,v 1.10 1998/09/29 22:44:03 lile Exp $
 *
 */

/*
 *
 * IBM Shared Ram and TROPIC Token-Ring
 *
 *   Token-Ring Network PC Adapter
 *   Token-Ring Network PC Adapter II
 *   Token-Ring Network Adapter/A
 *   Token-Ring Network 16/4 Adapter/A (long card)
 *   Token-Ring Network 16/4 Adapter
 *   Token-Ring Network 16/4 Adapter/A (short card)
 *   16/4 ISA-16 Adapter
 *   Auto 16/4 Token-Ring ISA Adapter
 *   Auto 16/4 Token-Ring MC Adapter <-- Not supported currently
 *
 * $Revision: 1.10 $
 * 
 */

/*
 * Programmed I/O Structures
 */

struct tok_pio {
   u_char  setup_read_1;  
   u_char  latch;        /* read     */
   u_char  release;      /* write    */
   u_char  setup_read_2; /* MCA Only */
   u_char  enable_int;   /* write    */
};

/*
 * Attachment Control Area Structures
 */

struct aca_reg {
    union {
        u_short r;
        struct {
             u_char e;
             u_char o;
        } b;
    } RRR;
#define rrr     RRR.r
#define rrr_e   RRR.b.e
#define rrr_o   RRR.b.o
    union {
        u_short r;
        struct {
             u_char e;
             u_char o;
        } b;
    } WRBR;
#define wrbr    WRBR.r
#define wrbr_e  WRBR.b.e
#define wrbr_o  WRBR.b.o
    union {
        u_short r;
        struct {
             u_char e;
             u_char o;
        } b;
    } WWOR;
#define wwor    WWOR.r
#define wwor_e  WWOR.b.e
#define wwor_o  WWOR.b.o
    union {
        u_short r;
        struct {
             u_char e;
             u_char o;
        } b;
    } WWCR;
#define wwcr    WWCR.r
#define wwcr_e  WWCR.b.e
#define wwcr_o  WWCR.b.o
    union {
        u_short r;
        struct {
             u_char e;
             u_char o;
        } b;
    } ISRP;
#define isrp    ISRP.r
#define isrp_e  ISRP.b.e
#define isrp_o  ISRP.b.o
    union {
        u_short r;
        struct {
             u_char e;
             u_char o;
        } b;
    } ISRA;
#define isra    ISRA.b
#define isra_e  ISRA.b.e
#define isra_o  ISRA.b.o
    union {
        u_short r;
        struct {
             u_char e;
             u_char o;
        } b;
    } TCR;
#define tcr     TCR.r
#define tcr_e   TCR.b.e
#define tcr_o   TCR.b.o
    union {
        u_short r;
        struct {
             u_char e;
             u_char o;
        } b;
   } TVR;
#define tvr     TVR.r
#define tvr_e   TVR.b.e
#define tvr_o   TVR.b.o
};

struct aca {
    struct  aca_reg  rw;
    u_char           fill0[16];
/*    union {
        u_short r;
        struct {
             u_char e;
             u_char o;
        } b;
    } SRPR;
#define srpr    SRPR.r
#define srpr_e  SRPR.b.e
#define srpr_o  SRPR.b.o
    u_char           fill1[14];
*/
    struct  aca_reg  reset;
    u_char           fill2[16];
    struct  aca_reg  set;
};

/*
 * PC to Adapter Commands ----------------
 */
#define DIR_CLOSE_ADAPTER           0x04
#define DIR_INTERRUPT               0x00
#define DIR_MODIFY_OPEN_PARMS       0x01
#define DIR_OPEN_ADAPTER            0x03
#define DIR_READ_LOG                0x08
#define DIR_RESTORE_OPEN_PARMS      0x02
#define DIR_SET_FUNCTIONAL_ADDRESS  0x07
#define DIR_SET_GROUP_ADDRESS       0x06

#define DLC_CLOSE_SAP               0x16
#define DLC_CLSE_STATION            0x1a
#define DLC_CONNECT_STATION         0x1b
#define DLC_FLOW_CONTROL            0x1d
#define DLC_MODIFY                  0x1d
#define DLC_OPEN_SAP                0x15
#define DLC_OPEN_STATION            0x19
#define DLC_REALLOCATE              0x17
#define DLC_RESET                   0x14
#define DLC_STATISTICS              0x1e

#define XMIT_DIR_FRAME              0x0a
#define XMIT_I_FRAME                0x0b
#define XMIT_UI_FRAME               0x0d
#define XMIT_XID_CMD                0x0e
#define XMIT_XID_RESP_FINAL         0x0f
#define XMIT_XID_RESP_NOT_FINAL     0x10
#define XMIT_TEST_CMD               0x11

/*
 * Adapter to PC Commands ---------------
 */
#define DLC_STATUS             0x83
#define RECEIVED_DATA          0x81
#define RING_STATUS_CHANGE     0x84
#define TRANSMIT_DATA_REQUEST  0x82

struct cb {                      /* Adapter command buffer locations */
    union  sr_srb *srb;
    struct sr_ssb *ssb;
    union  sr_arb *arb;
    union  sr_asb *asb;
};

struct log_data {                /* Adapter internal logs */
   u_char  line_errs;
   u_char  internal_errs;
   u_char  burst_errs;
   u_char  ac_errs;
   u_char  abort_delims;
   u_char  reserved;
   u_char  lost_frames;
   u_char  rec_congest_count;
   u_char  frame_copied_errs;
   u_char  freq_errs;
   u_char  token_errs;
   u_char  reserved1[3];
};

struct init_resp {                /* Initialization response */
   u_char  cmd;
   u_char  status;
   u_char  status_2;
   u_char  reserved[3];
   u_short bring_up_code;
   u_short encoded_addr;
   u_short level_addr;
   u_short adapter_addr;
   u_short parms_addr;
   u_short mac_addr;
   u_short util_addr2;
};

/* 
 * System Request Blocks (srb)
 */

union sr_srb {
    struct init_resp init_resp;
    /* Direct Interface Commands */
    struct {                /* dir.close.adapter (x04)*/
        u_char  cmd;
        u_char  reserved;
        u_char  retcode;
    }    close; 
    struct {                /* dir.interrupt.adapter (x00) */
        u_char  cmd;
        u_char  reserved;
        u_char  retcode;
    }    interrupt;
    struct {                /* dir.modify.open.parms (x01) */
        u_char  cmd;
        u_char  reserved;
        u_char  retcode;
        u_char  reserved1;
        u_short open_options;
    }    mod_params;
    struct {                /* dir.open.adapter (x03) */
        u_char  cmd;
        u_char  reserved[7];
        u_short open_options;
        u_char  node_address[6];
        u_char  group_address[4];
        u_char  funct_address[4];
        u_short num_rcv_buf;
        u_short rcv_buf_len;
        u_short dhb_length;
        u_char  num_dhb;
        u_char  reserved1;
        u_char  dlc_max_sap;
        u_char  dlc_max_sta;
        u_char  dlc_max_gsap;
        u_char  dlc_max_gmem;
        u_char  dlc_t1_tick_1;
        u_char  dlc_t2_tick_1;
        u_char  dlc_ti_tick_1;
        u_char  dlc_t1_tick_2;
        u_char  dlc_t2_tick_2;
        u_char  dlc_ti_tick_2;
        u_char  product_id[18];
    }    open_cmd;
    struct {                /* dir.open.adapter.resp (x03) */
        u_char  cmd;
        u_char  reserved;
        u_char  retcode;
        u_char  reserved1[3];
        u_char  open_error_code[2];
        u_short asb_address;
        u_short srb_address;
        u_short arb_address;
        u_short ssb_address;
    }    open_resp;
    struct {                /* dir.read.log (x08) */
        u_char  cmd;
        u_char  reserved;
        u_char  retcode;
        u_char  reserved1[3];
        struct  log_data log;
    }    read_log;
    struct {                /* dir.restore.open.parms (x02) */
        u_char  cmd;
        u_char  reserved;
        u_char  retcode;
        u_char  reserved1;
        u_short open_options;
    }    restore_parms;
    struct {                /* dir.set.funct.address (x07) */
        u_char  cmd;        /* dir.set.group.address (x06) */
        u_char  reserved;
        u_char  retcode;
        u_char  reserved1[3];
        u_char  address[4];
    }    set_addr;
    /* DLC adapter commands */
    struct {                /* dlc.close.sap (x16) */
        u_char  cmd;
        u_char  reserved;
        u_char  retcode;
        u_char  reserved1;
        u_char  station_id[2];
    }    close_sap;
    struct {                /* dlc.close.station (x1a) */
        u_char  cmd;
        u_char  reserved;
        u_char  retcode;
        u_char  reserved1;
        u_char  station_id[2];
    }    close_station;
    struct {                /* dlc.connect.station (x1b) */
        u_char  cmd;
        u_char  cmd_correlate;
        u_char  retcode;
        u_char  reserved;
        u_char  station_id[2];
        u_char  routing_info[18];
    }    connect_station;
    struct {                /* dlc.flow.control (x1d) */
        u_char  cmd;
        u_char  reserved;
        u_char  retcode;
        u_char  reserved1;
        u_char  station_id[2];
        u_char  flow_options;
    }    flow_control;
    struct {                /* dlc.modify (x1c) */
        u_char  cmd;
        u_char  reserved;
        u_char  retcode;
        u_char  reserved1;
        u_char  station_id[2];
        u_char  timer_t1;
        u_char  timer_t2;
        u_char  timer_ti;
        u_char  maxout;
        u_char  maxin;
        u_char  maxout_incr;
        u_char  max_retry_count;
        u_char  access_priority;
        u_char  sap_gsap_mem;
        u_char  gsaps[13];
    }    dlc_modify;
    struct {                /* dlc.open.sap (x15) */
        u_char  cmd;
        u_char  reserved;
        u_char  retcode;
        u_char  reserved1;
        u_char  station_id[2];
        u_char  timer_t1;
        u_char  timer_t2;
        u_char  timer_ti;
        u_char  maxout;
        u_char  maxin;
        u_char  maxout_incr;
        u_char  max_retry_count;
        u_char  gsap_max_mem;
        u_short max_i_field;
        u_char  sap_value;
        u_char  sap_options;
        u_char  station_count;
        u_char  sap_gsap_mem;
        u_char  gsaps[8];
    }    open_sap;
    struct {                /* dlc.open.station (x19) */
        u_char  cmd;
        u_char  reserved;
        u_char  retcode;
        u_char  reserved1;
        u_char  station_id[2];
        u_char  timer_t1;
        u_char  timer_t2;
        u_char  timer_ti;
        u_char  maxout;
        u_char  maxin;
        u_char  maxout_incr;
        u_char  max_retry_count;
        u_char  rsap_value;
        u_short max_i_field;
        u_char  station_options;
        u_char  reserved2;
        u_char  remote_address[6];
    }    open_station;
    struct {                /* dlc.rellocate (x17) */
        u_char  cmd;
        u_char  reserved;
        u_char  retcode;
        u_char  reserved1;
        u_char  station_id[2];
        u_char  option_byte;
        u_char  station_count;
        u_char  adapter_count;
        u_char  sap_count;
    }    reallocate;
    struct {                /* dlc.reset (x14) */
        u_char  cmd;
        u_char  reserved;
        u_char  retcode;
        u_char  station_id[2];
    }    dlc_reset;
    struct {                /* dlc_statistics (x1e) */
        u_char  cmd;
        u_char  reserved;
        u_char  retcode;
        u_char  reserved1;
        u_char  station_id[2];
        u_short counts_addr;
        u_short header_addr;
        u_char  header_length;
        u_char  reset_option;
    }    dlc_stats;
    struct {                /* transmit.XXX (x0a,x0b.x0d,x0e,x0f.x10.x11) */
        u_char  cmd;
        u_char  cmd_correlate;
        u_char  retcode;
        u_char  reserved1;
        u_char  station_id[2];
    }    xmit;
};

/*
 * System Status Blocks (ssb)
 */

struct sr_ssb {
    u_char  cmd;
    u_char  cmd_correlate;
    u_char  retcode;
    u_char  reserved;
    u_char  station_id[2];
    u_char  xmit_err;
};

/*
 * Adapter Request Block (arb)
 */

union sr_arb {
    struct {                /* dlc.status (x83) */
        u_char  cmd;
        u_char  reserved;
        u_char  station_id[2];
        u_short status;
        u_char  frmr_data[5];
        u_char  access_priority;
        u_char  remote_address[6];
        u_char  rsap_value;
    }    dlc_status;
    struct {                /* received_data (x81) */
        u_char  cmd;
        u_char  reserved[3];
        u_char  station_id[2];
        u_char receive_buffer[2];
        u_char  lan_hdr_length;
        u_char  dlc_hdr_length;
        u_short frame_length;
        u_char  ncb_type;
    }    rec_data;
    struct {                /* ring.status.change (x84) */
        u_char  cmd;
        u_char  reserved[5];
        u_short status;
    } ring_status_change;
    struct {                /* transmit.data.request (x82) */
        u_char  cmd;
        u_char  cmd_correlate;
        u_char  reserved[2];
        u_char  station_id[2];
        u_short dhb_address;
    }    xmit_request;
};

/*
 * Adapter Status Block
 */

union sr_asb {
    struct {                /* received.data (x81) */
        u_char  cmd;
        u_char  reserved;
        u_char  retcode;
        u_char  reserved1;
        u_char  station_id[2];
        u_char receive_buffer[2];
    }    rec_data;
    struct {                /* transmit.data.request */
        u_char  cmd;
        u_char  cmd_correlate;
        u_char  retcode;
        u_char  reserved;
        u_char  station_id[2];
        u_char  frame_length[2];
        u_char  header_length;
        u_char  rsap_value;
    }    data_req;
};

struct rec_buf {
    u_char  reserved[2];
    u_char  next_buffer[2];
    u_char  reserved1;
    u_char  receive_fs;
    u_short buffer_length;
    u_char  frame_data[256];
};
